
#include <stdarg.h>
#include <stdlib.h>
#include <iostream.h>


class Frame {
  int typeLen ;
  int rank ;
  int* shape ;
  int size ;
  char* data ;
public :
  void init(int, int, ...) ;
  void finalise() ;
  char* subs() ;
  char* subs(int, ...) ;
} ;



void Frame :: init(int _typeLen, int _rank, ...) {
  typeLen = _typeLen ;
  rank    = _rank ;
  shape   = new int [_rank] ;

  va_list ap ;

  va_start(ap, _rank) ;
  size = 1 ;
  for(int d = 0 ; d < _rank ; d++) {
    int n_d = va_arg(ap, int) ;

    shape [d] = n_d ;
    size *= n_d ;
  }
  va_end(ap) ;

  data = new char [size * _typeLen] ;
}


void Frame :: finalise() {
  delete data ;
  delete shape ;
}


char* Frame :: subs() {return data ; }

char* Frame :: subs(int i_0, ...) {
  va_list ap ;

  int stride, offset ;

  stride = size / shape [0] ;
  offset = i_0 * stride ;
  
  va_start(ap, n_0) ;
  for (int d = 1; d < rank ; d++) {
    int i_d = va_arg(ap, int) ;

    stride /= shape [d] ;
    offset += i_d * stride ;
  }
  va_end(ap) ;

  return data + typeLen * offset ;
}


// `Array' is a type-safe interface to the `Frame' class.

template<int R, class T>
class Array : private Frame {
  int shape [R] ;
  assertRank(int r) {if(r != R) exit(1) ; } ;
public :
  Array() {
    assertRank(0) ;
    init(sizeof(T), 0) ;
  }
  Array(int n_0) {
    assertRank(1) ;
    init(sizeof(T), 1, n_0) ;
  }
  Array(int n_0, int n_1) {
    assertRank(2) ;
    init(sizeof(T), 2, n_0, n_1) ;
  }
  Array(int n_0, int n_1, int n_2) {
    assertRank(3) ;
    init(sizeof(T), 3, n_0, n_1, n_2) ;
  }

  ~Array() {finalise() ; }

  T& operator()() {
    assertRank(0) ;
    return *((T*) subs()) ;
  } 
  T& operator()(int i_0) {
    assertRank(1) ;
    return *((T*) subs(i_0)) ;
  }
  T& operator()(int i_0, int i_1) {
    assertRank(2) ;
    return *((T*) subs(i_0, i_1)) ;
  }
  T& operator()(int i_0, int i_1, int i_2) {
    assertRank(3) ;
    return *((T*) subs(i_0, i_1, i_2)) ;
  }
} ;


main () {
  Array<3, float> a(5, 7, 11) ;
  Array<1, int> b(23) ;
  
  a(3, 2, 1) = 53.1 ;
  a(3, 2, 2) = 75.6 ;
  a(3, 3, 1) = 81.7 ;
  cout << a(3, 2, 1) << "\n" ;
  cout << a(3, 2, 2) << "\n" ;
  cout << a(3, 3, 1) << "\n" ;

  b(23) = a(3, 2, 1) ;
  cout << b(23) << "\n" ;
}

